#include <stdio.h>
#include <stdlib.h>

#include "tokenizer.h"

#define LOOKING_FOR_START_OF_TOKEN	1
#define LOOKING_FOR_END_OF_TOKEN	2

// prototypes of functions we do not necessarily want to put in tokenizer.h
// since we don't expect other programmers to use them directly 
// they are only useful for us to break down our job
void tokens_add(char** tokens, int maxnbtokens, char* user_input, int start, int stop); 

void tokenize(char** tokens, int maxnbtokens, char * str, int length)
{
		
	#ifdef DEBUG_MODE
	printf("\tTokenizing: %s\n", str); 
	#endif
	
	// let's start our search for tokens in the str string
	int 	start 	= -1; 
	int 	stop 	= -1; 
	int 	i 	= 0; 
	int	status	= LOOKING_FOR_START_OF_TOKEN; 
	
	do { 
	
		if ((str[i] == ' ') || (str[i] == '\t') || (str[i] == '\n') || (str[i] == '\0')){
			
			// we just detected a "separator" character
			
			if (status == LOOKING_FOR_START_OF_TOKEN){
				// we are looking for a start of token
				// this won't be here
			} else if (status == LOOKING_FOR_END_OF_TOKEN){
				// we already noticed the index where a token started
				// it just ended 
				stop = i -1; 
				
				// let's process the token we just identified
				tokens_add(tokens, maxnbtokens, str, start, stop);
				
				// we will now keep looking for the next token
				status = LOOKING_FOR_START_OF_TOKEN; 
			}
			
		} else {
		
			// we are on a character that is part of a token
			if (status == LOOKING_FOR_START_OF_TOKEN){
				// we are looking for a start of token
				// this is it
				start = i; 
				
				// now, let's keep looking at other characters 
				// until we encounter a "separator" 
				// that will tell us we found the end of this token
				status = LOOKING_FOR_END_OF_TOKEN; 
			} else if (status == LOOKING_FOR_END_OF_TOKEN){
				// we already noticed the index where a token started
				// this character is just part of it

			}
		
		}
		
		i++; 
	
	} while (i < length); 
	#ifdef DEBUG_MODE
	printf("\tDONE tokenizing %s\n", str); 
	#endif
}



void tokens_display(char** tokens, int maxnbtokens)
{
	printf("\tIdentified Tokens:\n"); 
	int i; 
	for(i=0 ; i < maxnbtokens ; i++)
		if (tokens[i] != NULL)
			printf("\tTokens[%d] \t= [%s]\n", i, tokens[i]); 
		else
			return; 
}



char** tokens_allocate(int maxnbtokens)
{
	// let's allocate the data structure to contain all our tokens copies
		
	char** tmp = malloc(sizeof(char*) * maxnbtokens); 
	
	if (tmp == NULL){
		perror("Couldn't allocate tokens array"); 
		exit(-1); 
	}
	
	// we have now an array holding n pointers on strings
	// each time we discover a new token we will allocate a string 
	// to hold a copy of it and put its address at the next available 
	// pointer in strings_array. To make it easy to identify which pointer(s)
	// are free, we will set them all to NULL initially
	
	int n; 
	for(n=0 ; n < maxnbtokens; n++)
		tmp[n] = NULL; 
	
	return tmp; 
}

void tokens_deallocate(char** tokens, int maxnbtokens)
{
	#ifdef DEBUG_MODE
	printf("\tDeallocating tokens...\n"); fflush(stdout); 
	#endif 
	
	int i; 
	for(i=0 ; i < maxnbtokens ; i++)
		if (tokens[i] != NULL) 	free(tokens[i]); 
		else			break;
	
	free(tokens); 
	
	#ifdef DEBUG_MODE
	printf("\tDONE deallocating tokens\n"); fflush(stdout);
	#endif
}


void tokens_add(char** tokens, int maxnbtokens, 
		char* user_input, 
		int start, int stop)
{
	int n = 0; 
	
	// look for the first unused pointer in strings_array 
	for ( n = 0 ; (n < maxnbtokens) && (tokens[n] != NULL) ; n++); 
	
	if (n == maxnbtokens){ 
		printf("Exhausted max number of tokens (%d) in tokens_add function\n", n);
		fflush(stdout); 
		exit(-1); 
	}
	
	// let's measure the length of the token we found in user_input
	int token_size  = stop - start + 1 + 1; // for the '\0' we'll have to add

	// allocating a dynamical string to hold this token's copy
	char* token_itself = malloc(sizeof(char) * token_size); 
	
	//let's make a copy of the token we found in the user_input string
	// we copy all the characters between start and stop (both inclusive)
	strncpy(token_itself, user_input + start, token_size-1); 
	
	// then we add a '\0' to make sure our newly created strings is properly
	// formatted
	token_itself[token_size-1] = '\0'; 
	
	// we add this new string to the strings_array
	tokens[n] = token_itself; 
	
	#ifdef DEBUG_MODE
	printf("\tJust adding token at index %d start = %d stop = %d  [%s] ...\n", n, start, stop, tokens[n]); 
	#endif
	
}
